from __future__ import annotations

import itertools
from typing import Any
from typing import TypeVar

import attrs
import pytest
from lsprotocol import types
from pygls import IS_WIN
from pygls.workspace import Workspace

from esbonio.server import EsbonioLanguageServer
from esbonio.server import Uri
from esbonio.server._configuration import _merge_configs
from esbonio.server._configuration import _uri_to_scope

T = TypeVar("T")


@attrs.define
class ExampleConfig:
    log_level: str = "error"
    log_names: list[str] = attrs.field(factory=list)


@attrs.define
class ColorConfig:
    color: str
    scope: str = attrs.field(default="${scope}")


@pytest.fixture
def server():
    """Return a server instance for testing."""
    _server = EsbonioLanguageServer()
    return _server


@pytest.mark.parametrize(
    "init_options,workspace_config,file_config,section,spec,scope,expected",
    [
        pytest.param(  # When there's no configuration at all, just return the defaults.
            {},
            {},
            {},
            "esbonio.server",
            ExampleConfig,
            None,
            ExampleConfig(log_level="error", log_names=[]),
            id="no-config",
        ),
        pytest.param(
            dict(esbonio=dict(server=dict(logLevel="info"))),
            {},
            {},
            "esbonio.server",
            ExampleConfig,
            None,
            ExampleConfig(log_level="info", log_names=[]),
            id="init-options",
        ),
        pytest.param(  # Handle the case where the client omits the top-level namespace
            dict(server=dict(logLevel="info")),
            {},
            {},
            "esbonio.server",
            ExampleConfig,
            None,
            ExampleConfig(log_level="info", log_names=[]),
            id="init-options-no-namespace",
        ),
        pytest.param(
            {},
            {"": dict(esbonio=dict(server=dict(logLevel="info")))},
            {},
            "esbonio.server",
            ExampleConfig,
            None,
            ExampleConfig(log_level="info", log_names=[]),
            id="global-workspace-config",
        ),
        pytest.param(
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            {},
            "esbonio.server",
            ExampleConfig,
            "file:///path/to/workspace/file.txt",
            ExampleConfig(log_level="info", log_names=[]),
            id="workspace-config[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            {},
            "esbonio.server",
            ExampleConfig,
            "file:///c:/path/to/workspace/file.txt",
            ExampleConfig(log_level="info", log_names=[]),
            id="workspace-config[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known workspace
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            {},
            "esbonio.server",
            ExampleConfig,
            "file:///path/to/other/workspace/file.txt",
            ExampleConfig(log_level="error", log_names=[]),
            id="workspace-config-scope-mismatch[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known workspace
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            {},
            "esbonio.server",
            ExampleConfig,
            "file:///c:/path/to/other/workspace/file.txt",
            ExampleConfig(log_level="error", log_names=[]),
            id="workspace-config-scope-mismatch[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known file
            {},
            {},
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///path/to/workspace/docs/file.txt",
            ExampleConfig(log_level="info", log_names=[]),
            id="file-config[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known file
            {},
            {},
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ExampleConfig(log_level="info", log_names=[]),
            id="file-config[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known file
            {},
            {},
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///path/to/workspace/workspace/file.txt",
            ExampleConfig(log_level="error", log_names=[]),
            id="file-config-scope-mismatch[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Handle the case where the scope does not match a known file
            {},
            {},
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info"))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///c:/path/to/workspace/workspace/file.txt",
            ExampleConfig(log_level="error", log_names=[]),
            id="file-config-scope-mismatch[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Allow the client to override the file based config
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="debug"))
                )
            },
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info", logNames=["file"]))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///path/to/workspace/docs/file.txt",
            ExampleConfig(log_level="debug", log_names=["file"]),
            id="workspace-file-override[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Allow the client to override the file based config
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(server=dict(logLevel="debug"))
                )
            },
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(server=dict(logLevel="info", logNames=["file"]))
                )
            },
            "esbonio.server",
            ExampleConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ExampleConfig(log_level="debug", log_names=["file"]),
            id="workspace-file-override[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Check that we can expand config variables correctly
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {},
            "esbonio.colors",
            ColorConfig,
            "file:///path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///path/to/workspace"),
            id="scope-variable[workspace][unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Check that we can expand config variables correctly
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {},
            "esbonio.colors",
            ColorConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///c%3A/path/to/workspace"),
            id="scope-variable[workspace][win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # Check that we can expand config variables correctly
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///path/to/workspace/docs"),
            id="scope-variable[workspace+file][unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Check that we can expand config variables correctly
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///c%3A/path/to/workspace/docs"),
            id="scope-variable[workspace+file][win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(  # The user should still be able to override them
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="file:///my/scope"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///my/scope"),
            id="scope-variable-override[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # The user should still be able to override them
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="file:///my/scope"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="file:///my/scope"),
            id="scope-variable-override[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="${scopePath}"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="/path/to/workspace/docs"),
            id="scope-path-variable[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="${scopePath}"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="/c:/path/to/workspace/docs"),
            id="scope-path-variable[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            {},
            {
                "file:///path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="${scopeFsPath}"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="/path/to/workspace/docs"),
            id="scope-fspath-variable[unix]",
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            {},
            {
                "file:///c%3A/path/to/workspace": dict(
                    esbonio=dict(colors=dict(color="red"))
                ),
            },
            {
                "file:///c%3A/path/to/workspace/docs": dict(
                    esbonio=dict(colors=dict(color="blue", scope="${scopeFsPath}"))
                ),
            },
            "esbonio.colors",
            ColorConfig,
            "file:///c:/path/to/workspace/docs/file.txt",
            ColorConfig(color="red", scope="c:\\path\\to\\workspace\\docs"),
            id="scope-fspath-variable[win]",
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
    ],
)
def test_get_configuration(
    server: EsbonioLanguageServer,
    init_options: dict[str, Any],
    workspace_config: dict[str, Any],
    file_config: dict[str, Any],
    section: str,
    spec: type[T],
    scope: str | None,
    expected: T,
):
    """Ensure that we can get configuration values correctly.

    The number of parameters to this test case is a lot! But the test itself is fairly
    simple. Given the current configuration of the server, can we resolve the
    configuration correctly and return the expected set of values?

    Parameters
    ----------
    init_options
       The server's initialization options.

    workspace_config
       The server's cached ``workspace/configuration`` requests

    file_config
       The server's cached config coming from configuration files

    section
       The configuration section to retrieve

    spec
       The datatype describing the configuration

    scope
       The scope at which to get the configuration

    expected
       The expected value.
    """
    server.configuration.initialization_options = init_options
    server.configuration._workspace_config = workspace_config
    server.configuration._file_config = file_config

    # Infer the workspace based on the given config.
    workspace_folders = [
        types.WorkspaceFolder(uri=uri, name=f"workspace-{idx}")
        for idx, uri in enumerate(workspace_config.keys())
        if uri != ""
    ]
    server.protocol._workspace = Workspace(None, workspace_folders=workspace_folders)

    scope_uri = Uri.parse(scope) if scope else None

    actual = server.configuration.get(section, spec, scope=scope_uri)
    assert actual == expected


@pytest.mark.parametrize(
    "known_scopes,setup",
    [
        *itertools.product(
            [
                ["file:///path/to/test"],
            ],
            [
                ("file:///path/to/test/file.txt", "file:///path/to/test"),
                ("file:///path/to/other/file.txt", ""),
                (None, ""),
            ],
        ),
        *itertools.product(
            [
                ["file:///path/to/test", "file:///path/to/test/sub/folder"],
            ],
            [
                ("file:///path/to/test/file.txt", "file:///path/to/test"),
                ("file:///path/to/test/sub/file.txt", "file:///path/to/test"),
                (
                    "file:///path/to/test/sub/folder/file.txt",
                    "file:///path/to/test/sub/folder",
                ),
            ],
        ),
    ],
)
@pytest.mark.skipif(IS_WIN, reason="windows")
def test_uri_to_scope(known_scopes, setup):
    """Ensure that we can determine the correct scope for the given uri."""

    uri, expected = setup
    parsed_uri = Uri.parse(uri) if uri else None
    scope = _uri_to_scope(known_scopes, parsed_uri)
    assert scope == expected


@pytest.mark.parametrize(
    "known_scopes,setup",
    [
        *itertools.product(
            [
                ["file:///c:/path/to/test", "file:///c:/path/to/test/sub/folder"],
            ],
            [
                ("file:///c:/path/to/test/file.txt", "file:///c%3A/path/to/test"),
                ("file:///c%3A/path/to/test/file.txt", "file:///c%3A/path/to/test"),
                ("file:///C:/path/to/test/file.txt", "file:///c%3A/path/to/test"),
                ("file:///C%3A/path/to/test/file.txt", "file:///c%3A/path/to/test"),
                ("file:///c:/path/to/test/sub/file.txt", "file:///c%3A/path/to/test"),
                (
                    "file:///c:/path/to/test/sub/folder/file.txt",
                    "file:///c%3A/path/to/test/sub/folder",
                ),
            ],
        ),
    ],
)
@pytest.mark.skipif(not IS_WIN, reason="windows only")
def test_uri_to_scope_windows(known_scopes, setup):
    """Ensure that we can determine the correct scope for the given uri."""

    uri, expected = setup
    scope = _uri_to_scope(known_scopes, Uri.parse(uri))
    assert scope == expected


@pytest.mark.parametrize(
    "configs, expected",
    [
        (  # Last value wins
            [
                {"a": 1},
                {"a": 3},
                {"a": 2},
            ],
            {"a": 2},
        ),
        (  # None is a valid value
            [
                {"a": 1},
                {"a": 2},
                {"a": None},
            ],
            {"a": None},
        ),
        (  # Simple merge
            [
                {"a": 1},
                {"b": 3},
                {"c": 2},
            ],
            {"a": 1, "b": 3, "c": 2},
        ),
        (  # Ensure nested values are merged as well
            [
                {"a": 1, "b": {"c": 1}},
                {"b": {"c": 4}},
                {"b": {"d": 7}},
            ],
            {"a": 1, "b": {"c": 4, "d": 7}},
        ),
        (  # Handle the case where there's a type mismatch.
            [
                {"a": {"c": 1}},
                {"a": 2},
                {"a": {"d": 7}},
            ],
            {"a": {"c": 1, "d": 7}},
        ),
        (  # Let the final config dictate the type of a given field.
            [
                {"a": {"c": 1}},
                {"a": {"d": 7}},
                {"a": 2},
            ],
            {"a": 2},
        ),
    ],
)
def test_merge_configs(configs, expected):
    """Ensure that we can merge configurations together correctly."""
    assert expected == _merge_configs(*configs)



"""Helper functions for completion support"""

from __future__ import annotations

import re
import typing

from lsprotocol import types

from esbonio import server

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from esbonio.sphinx_agent.types import Directive

    DirectiveRenderer = Callable[
        [server.CompletionContext, Directive], types.CompletionItem | None
    ]

    DirectiveArgumentRenderer = Callable[
        [server.CompletionContext, types.CompletionItem], types.CompletionItem | None
    ]


WORD = re.compile("[a-zA-Z]+")
_DIRECTIVE_RENDERERS: dict[tuple[str, str], DirectiveRenderer] = {}
"""CompletionItem rendering functions for directives."""

_DIRECTIVE_ARGUMENT_RENDERERS: dict[tuple[str, str], DirectiveArgumentRenderer] = {}
"""CompletionItem rendering functions for role targets."""


def directive_renderer(*, language: str, insert_behavior: str):
    """Define a new rendering function."""

    def fn(f: DirectiveRenderer) -> DirectiveRenderer:
        _DIRECTIVE_RENDERERS[(language, insert_behavior)] = f
        return f

    return fn


def directive_argument_renderer(*, language: str, insert_behavior: str):
    """Define a new rendering function."""

    def fn(f: DirectiveArgumentRenderer) -> DirectiveArgumentRenderer:
        _DIRECTIVE_ARGUMENT_RENDERERS[(language, insert_behavior)] = f
        return f

    return fn


def get_directive_renderer(
    language: str, insert_behavior: str
) -> DirectiveRenderer | None:
    """Return the directive renderer to use.

    Parameters
    ----------
    language
       The source language the completion item will be inserted into

    insert_behavior
       How the completion should behave when inserted.

    Returns
    -------
    Optional[DirectiveRenderer]
       The rendering function to use that matches the given criteria, if available.
    """
    return _DIRECTIVE_RENDERERS.get((language, insert_behavior), None)


def get_directive_argument_renderer(
    language: str, insert_behavior: str
) -> DirectiveArgumentRenderer | None:
    """Return the directive argument renderer to use.

    Parameters
    ----------
    language
       The source language the completion item will be inserted into

    insert_behavior
       How the completion should behave when inserted.

    Returns
    -------
    Optional[DirectiveArgumentRenderer]
       The rendering function to use that matches the given criteria, if available.
    """
    return _DIRECTIVE_ARGUMENT_RENDERERS.get((language, insert_behavior), None)


@directive_renderer(language="rst", insert_behavior="insert")
def render_rst_directive_with_insert_text(
    context: server.CompletionContext,
    directive: Directive,
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``insertText`` fields.

    This implements the ``insert`` behavior for directives.
    Parameters
    ----------
    context
       The context in which the completion is being generated.

    directive
       The directive.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """
    insert_text = f".. {directive.name}::"
    user_text = context.match.group(0).strip()

    # Since we can't replace any existing text, it only makes sense
    # to offer completions that align with what the user has already written.
    if not insert_text.startswith(user_text):
        return None

    # Except that's not entirely true... to quote the LSP spec. (emphasis added)
    #
    # > in the model the client should filter against what the user has already typed
    # > **using the word boundary rules of the language** (e.g. resolving the word
    # > under the cursor position). The reason for this mode is that it makes it
    # > extremely easy for a server to implement a basic completion list and get it
    # > filtered on the client.
    #
    # So in other words... if the cursor is inside a word, that entire word will be
    # replaced with what we have in `insert_text` so we need to be able to do something
    # like
    #    ..         -> image::
    #    .. im      -> image::
    #
    #    ..         -> code-block::
    #    .. cod     -> code-block::
    #    .. code-bl -> block::
    #
    #    ..         -> c:function::
    #    .. c       -> c:function::
    #    .. c:      -> function::
    #    .. c:fun   -> function::
    #
    # And since the client is free to interpret this how it likes, it's unlikely we'll
    # be able to get this right in all cases for all clients. So for now this is going
    # to target Kate's interpretation since it currently does not support ``text_edit``
    # and it was the editor that prompted this to be implemented in the first place.
    #
    # See: https://github.com/swyddfa/esbonio/issues/471

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", " ")):
        start_index = len(user_text)

    # Look for groups of word chars, replace text until the start of the final group
    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_directive_common(directive)
    item.insert_text = insert_text[start_index:]
    item.insert_text_format = types.InsertTextFormat.PlainText
    return item


@directive_argument_renderer(language="rst", insert_behavior="replace")
def render_rst_argument_with_text_edit(
    context: server.CompletionContext, item: types.CompletionItem
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``textEdit``.

    This implements the ``replace`` insert behavior for rst directive arguments.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    item
       The ``CompletionItem`` representing the directive argument.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the item should be skipped
    """
    return item


@directive_renderer(language="rst", insert_behavior="replace")
def render_rst_directive_with_text_edit(
    context: server.CompletionContext,
    directive: Directive,
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` for a reStructuredText directive using the
    ``textEdit`` field.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    directive
       The directive to render.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """
    match = context.match

    # Calculate the range of text the CompletionItems should edit, we don't need to
    # touch indentation.
    start = match.span()[0] + match.group(0).find(".")

    if match.group("argument"):
        # If there is an existing argument to the directive, we should leave it
        # untouched
        end = match.span()[0] + match.group(0).find("::") + 2
    else:
        end = match.span()[1]

    insert_text = f".. {directive.name}::"

    item = _render_directive_common(directive)
    item.filter_text = insert_text
    item.insert_text_format = types.InsertTextFormat.PlainText
    item.text_edit = types.TextEdit(
        new_text=insert_text,
        range=types.Range(
            start=types.Position(line=context.position.line, character=start),
            end=types.Position(line=context.position.line, character=end),
        ),
    )

    return item


@directive_renderer(language="markdown", insert_behavior="replace")
def render_myst_directive_with_text_edit(
    context: server.CompletionContext,
    directive: Directive,
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` for a MyST directive using the ``textEdit`` field.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    directive
       The directive to render.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item
    """

    # Calculate the range of text the CompletionItems should edit, we don't need to
    # touch indentation.
    start = context.match.span()[0] + context.match.group(0).find("`")

    if has_argument := context.match.group("argument"):
        end = context.match.span()[0] + context.match.group(0).find("}") + 1
    else:
        end = context.match.span()[1]

    if context.snippet_support and not has_argument:
        insert_text = f"```{{{directive.name}}} $0\n```"
        insert_text_format = types.InsertTextFormat.Snippet
    else:
        insert_text = f"```{{{directive.name}}}"
        insert_text_format = types.InsertTextFormat.PlainText

    item = _render_directive_common(directive)
    item.filter_text = insert_text
    item.insert_text_format = insert_text_format
    item.text_edit = types.TextEdit(
        new_text=insert_text,
        range=types.Range(
            start=types.Position(line=context.position.line, character=start),
            end=types.Position(line=context.position.line, character=end),
        ),
    )

    return item


@directive_argument_renderer(language="markdown", insert_behavior="replace")
def render_myst_argument_with_text_edit(
    context: server.CompletionContext, item: types.CompletionItem
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``textEdit``.

    This implements the ``replace`` insert behavior for MyST directive arguments.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    item
       The ``CompletionItem`` representing the directive argument.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the item should be skipped
    """
    return item


@directive_renderer(language="markdown", insert_behavior="insert")
def render_myst_directive_with_insert_text(
    context: server.CompletionContext,
    directive: Directive,
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` for a MyST directive using the ``insertText`` field.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    directive
       The directive to render.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item
    """
    if context.snippet_support:
        insert_text = f"```{{{directive.name}}} $0\n```"
        insert_text_format = types.InsertTextFormat.Snippet
    else:
        insert_text = f"```{{{directive.name}}}"
        insert_text_format = types.InsertTextFormat.PlainText

    user_text = context.match.group(0).strip()

    # Since we can't replace any existing text, it only makes sense
    # to offer completions that align with what the user has already written.
    if not insert_text.startswith(user_text):
        return None

    # See comment in `render_rst_directive_with_insert_text` above for an extended
    # discussion on how `insertText` completions work.

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", "{", "}", "`")):
        start_index = len(user_text)

    else:
        # Look for groups of word chars, replace text until the start of the final group
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_directive_common(directive)
    item.insert_text = insert_text[start_index:]
    item.insert_text_format = insert_text_format

    return item


def _render_directive_common(directive: Directive) -> types.CompletionItem:
    """Render the common fields of a directive's completion item."""

    return types.CompletionItem(
        label=directive.name,
        detail=directive.implementation,
        kind=types.CompletionItemKind.Class,
        data={"completion_type": "directive"},
    )


# def _render_directive_option_with_insert_text(
#     context: CompletionContext,
#     directive: Directive,
# ) -> Optional[types.CompletionItem]:
#     """Render a directive option's ``CompletionItem`` using the ``insertText`` field.

#     This implements the ``insert`` insert behavior for directive options.

#     Parameters
#     ----------
#     context
#        The context in which the completion is being generated.

#     name
#        The name of the directive option, as it appears in an rst file.

#     directive
#        The name of the directive, as it appears in an rst file.

#     implementation
#        The class implementing the directive.

#     """

#     insert_text = f":{name}:"
#     user_text = context.match.group(0).strip()

#     if not insert_text.startswith(user_text):
#         return None

#     if user_text.endswith((":", "-", " ")):
#         start_index = len(user_text)

#     else:
#         start_indices = [m.start() for m in WORD.finditer(user_text)] or [
#             len(user_text)
#         ]
#         start_index = max(start_indices)

#     item = _render_directive_option_common(name, directive, implementation)
#     item.insert_text = insert_text[start_index:]
#     return item


# def _render_directive_option_with_text_edit(
#     context: CompletionContext,
#     name: str,
#     directive: str,
#     implementation: Type[Directive],
# ) -> CompletionItem:
#     """Render a directive option's ``CompletionItem`` using the``textEdit`` field.

#     This implements the ``replace`` insert behavior for directive options.

#     Parameters
#     ----------
#     context
#        The context in which the completion is being generated.

#     name
#        The name of the directive option, as it appears in an rst file.

#     directive
#        The name of the directive, as it appears in an rst file.

#     implementation
#        The class implementing the directive.

#     """

#     match = context.match
#     groups = match.groupdict()

#     option = groups["option"]
#     start = match.span()[0] + match.group(0).find(option)
#     end = start + len(option)

#     range_ = Range(
#         start=Position(line=context.position.line, character=start),
#         end=Position(line=context.position.line, character=end),
#     )

#     insert_text = f":{name}:"

#     item = _render_directive_option_common(name, directive, implementation)
#     item.filter_text = insert_text
#     item.text_edit = TextEdit(range=range_, new_text=insert_text)

#     return item


# def _render_directive_option_common(
#     name: str, directive: str, impl: Type[Directive]
# ) -> CompletionItem:
#     """Render the common fields of a directive option's completion item."""

#     try:
#         impl_name = f"{impl.__module__}.{impl.__name__}"
#     except AttributeError:
#         impl_name = f"{impl.__module__}.{impl.__class__.__name__}"

#     return CompletionItem(
#         label=name,
#         detail=f"{impl_name}:{name}",
#         kind=CompletionItemKind.Field,
#         data={"completion_type": "directive_option", "for_directive": directive},
#     )



from __future__ import annotations

import pathlib

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

EXPECTED = {
    "function",
    "module",
    "option",
    "program",
    "image",
    "toctree",
    "c:macro",
    "c:function",
    "py:function",
    "py:module",
    "std:program",
    "std:option",
}

RST_EXPECTED = EXPECTED.copy()
MYST_EXPECTED = {"eval-rst", *EXPECTED}

UNEXPECTED = {
    "macro",
    "restructuredtext-test-directive",
}

RST_UNEXPECTED = {"eval-rst", *UNEXPECTED}
MYST_UNEXPECTED = UNEXPECTED.copy()

# Code blocks
LEXERS = {"python", "python-console", "nix"}

# Filepaths
ROOT_FILES = {"conf.py", "index.rst", "myst", "rst"}
RST_FILES = {"directives.rst", "roles.rst", "domains"}
MYST_FILES = {"directives.md", "roles.md"}


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        # Test cases covering directive name completion
        (".", None, None),
        ("..", RST_EXPECTED, RST_UNEXPECTED),
        (".. ", RST_EXPECTED, RST_UNEXPECTED),
        (".. d", RST_EXPECTED, RST_UNEXPECTED),
        (".. code-b", RST_EXPECTED, RST_UNEXPECTED),
        (".. codex-block:: ", None, None),
        (".. c:", RST_EXPECTED, RST_UNEXPECTED),
        (".. _some_label:", None, None),
        ("   .", None, None),
        ("   ..", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. ", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. d", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. doctest:: ", None, None),
        ("   .. code-b", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. codex-block:: ", None, None),
        ("   .. _some_label:", None, None),
        ("   .. c:", RST_EXPECTED, RST_UNEXPECTED),
        # Test cases covering directive argument completion for...
        #
        # -- pygments lexers
        (".. code-block:: ", LEXERS, None),
        (".. highlight:: ", LEXERS, None),
        (".. sourcecode:: ", LEXERS, None),
        # -- filepaths
        (".. image:: /", ROOT_FILES, None),
        (".. image:: ../", ROOT_FILES, None),
        (".. image:: ", RST_FILES, None),
        (".. image:: .", RST_FILES, None),
        (".. image:: ./", RST_FILES, None),
        (".. figure:: ./", RST_FILES, None),
        (".. include:: ./", RST_FILES, None),
        (".. literalinclude:: ./", RST_FILES, None),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_rst_directive_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer directive completions in rst
    documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "directives.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your note here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        # Test cases covering directive name completions
        ("`", None, None),
        ("``", None, None),
        # -- Unless the user types a '{', we should suggest languauge names
        ("```", LEXERS, MYST_EXPECTED | MYST_UNEXPECTED),
        ("```{", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{d", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{code-b", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{codex-block} ", None, None),
        ("```{c:", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   `", None, None),
        ("   ``", None, None),
        # -- Unless the user types a '{', we should suggest languauge names
        ("   ```", LEXERS, MYST_EXPECTED | MYST_UNEXPECTED),
        ("   ```{", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{d", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{doctest}", None, None),
        ("   ```{code-b", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{codex-block}", None, None),
        ("   ```{c:", MYST_EXPECTED, MYST_UNEXPECTED),
        # Test cases covering directive argument completions for...
        #
        # -- pygments lexers
        ("```{code-block} ", LEXERS, None),
        ("```{highlight} ", LEXERS, None),
        ("```{sourcecode} ", LEXERS, None),
        # -- filepaths
        ("```{image} /", ROOT_FILES, None),
        ("```{image} ../", ROOT_FILES, None),
        ("```{image} ", MYST_FILES, None),
        ("```{image} .", MYST_FILES, None),
        ("```{image} ./", MYST_FILES, None),
        ("```{figure} ./", MYST_FILES, None),
        ("```{include} ./", MYST_FILES, None),
        ("```{literalinclude} ./", MYST_FILES, None),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_myst_directive_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "directives.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your note here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.parametrize(
    "filename,expected",
    [
        (
            ["workspaces", "demo", "rst", "directives.rst"],
            [
                types.DocumentLink(
                    target="${ROOT}/rst/directives.rst",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=49, character=18),
                        end=types.Position(line=49, character=34),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/rst/symbols.rst",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=63, character=18),
                        end=types.Position(line=63, character=31),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html#rst-directives",
                    tooltip="Directives - Sphinx v",  # don't check for a precise version!
                    range=types.Range(
                        start=types.Position(line=3, character=121),
                        end=types.Position(line=3, character=135),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-literalinclude",
                    tooltip="literalinclude - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=21, character=28),
                        end=types.Position(line=21, character=42),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-code",
                    tooltip="code - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=30, character=28),
                        end=types.Position(line=30, character=32),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-code-block",
                    tooltip="code-block - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=31, character=28),
                        end=types.Position(line=31, character=38),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-highlight",
                    tooltip="highlight - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=32, character=28),
                        end=types.Position(line=32, character=37),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-sourcecode",
                    tooltip="sourcecode - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=33, character=28),
                        end=types.Position(line=33, character=38),
                    ),
                ),
            ],
        ),
        (
            ["workspaces", "demo", "myst", "directives.md"],
            [
                types.DocumentLink(
                    target="${ROOT}/myst/directives.md",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=47, character=15),
                        end=types.Position(line=47, character=30),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/symbols.md",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=59, character=15),
                        end=types.Position(line=59, character=27),
                    ),
                ),
                types.DocumentLink(
                    target="https://myst-parser.readthedocs.io/en/latest/syntax/roles-and-directives.html#syntax-directives",
                    tooltip="Directives - a block-level extension point - MyST Parser v",
                    range=types.Range(
                        start=types.Position(line=2, character=126),
                        end=types.Position(line=2, character=143),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-literalinclude",
                    tooltip="literalinclude - Sphinx v",  # don't check for a precise version!
                    range=types.Range(
                        start=types.Position(line=19, character=28),
                        end=types.Position(line=19, character=42),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-code",
                    tooltip="code - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=28, character=28),
                        end=types.Position(line=28, character=32),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-code-block",
                    tooltip="code-block - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=29, character=28),
                        end=types.Position(line=29, character=38),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-highlight",
                    tooltip="highlight - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=30, character=28),
                        end=types.Position(line=30, character=37),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-sourcecode",
                    tooltip="sourcecode - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=31, character=28),
                        end=types.Position(line=31, character=38),
                    ),
                ),
            ],
        ),
    ],
)
async def test_directive_document_links(
    client: LanguageClient,
    uri_for,
    filename: list[str],
    expected: list[types.DocumentLink],
):
    """Ensure that we handle ``textDocument/documentLink`` requests correctly."""

    root_uri = str(uri_for("workspaces", "demo"))
    test_uri = uri_for(*filename)

    links = await client.text_document_document_link_async(
        types.DocumentLinkParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    assert len(links) == len(expected)

    for link, actual in zip(expected, links, strict=False):
        assert link.range == actual.range

        target = link.target.replace("${ROOT}", root_uri)
        assert target == actual.target

        if link.tooltip is not None:
            assert actual.tooltip.startswith(link.tooltip)


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.parametrize(
    "filename,position,expected",
    [
        (
            ["workspaces", "demo", "rst", "directives.rst"],
            # Requests for the diretive itself should return nothing
            types.Position(line=63, character=12),
            None,
        ),
        (
            ["workspaces", "demo", "rst", "directives.rst"],
            types.Position(line=63, character=22),
            [
                types.Location(
                    uri="${ROOT}/rst/symbols.rst",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "myst", "directives.md"],
            # Requests for the diretive itself should return nothing
            types.Position(line=59, character=11),
            None,
        ),
        (
            ["workspaces", "demo", "myst", "directives.md"],
            types.Position(line=59, character=21),
            [
                types.Location(
                    uri="${ROOT}/myst/symbols.md",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
    ],
)
async def test_directive_argument_definitions(
    client: LanguageClient,
    uri_for,
    filename: list[str],
    position: types.Position,
    expected: list[types.Location] | None,
):
    """Ensure that we handle ``textDocument/definition`` requests correctly for
    directive arguments."""

    root_uri = str(uri_for("workspaces", "demo"))
    test_uri = uri_for(*filename)

    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()

    # Open the file - needed so that esbonio can correctly determine the language_id of
    # the document. Strictly speaking, you could probably consider the fact that this is
    # necessary to be a bug... but 99% of the time I'm fairly sure the client will have
    # done this before making the goto definition call.
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=str(test_uri),
                language_id="restructuredtext"
                if fpath.suffix == ".rst"
                else "markdown",
                version=1,
                text=contents,
            )
        )
    )

    definitions = await client.text_document_definition_async(
        types.DefinitionParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri)),
            position=position,
        )
    )

    if expected is None:
        assert definitions is None

    else:
        assert len(definitions) == len(expected)

        location: types.Location
        for location, actual in zip(expected, definitions, strict=False):
            expected_uri = location.uri.replace("${ROOT}", root_uri)
            assert expected_uri == actual.uri
            assert location.range == actual.range



from sphinx.config import Config

from ..app import Database
from ..app import Sphinx
from ..types import Uri
from ..util import as_json

DIAGNOSTICS_TABLE = Database.Table(
    "diagnostics",
    [
        Database.Column(name="uri", dtype="TEXT"),
        Database.Column(name="diagnostic", dtype="JSON"),
    ],
)


def init_db(app: Sphinx, config: Config):
    app.esbonio.db.ensure_table(DIAGNOSTICS_TABLE)


def clear_diagnostics(app: Sphinx, docname: str, source):
    """Clear the diagnostics assocated with the given file."""
    uri = Uri.for_file(app.env.doc2path(docname, base=True))
    app.esbonio.diagnostics.pop(uri, None)


def sync_diagnostics(app: Sphinx, *args):
    app.esbonio.db.clear_table(DIAGNOSTICS_TABLE)

    results = []
    diagnostics = app.esbonio.diagnostics

    for uri, items in diagnostics.items():
        for item in items:
            results.append((str(uri), as_json(item)))

    app.esbonio.db.insert_values(DIAGNOSTICS_TABLE, results)


def setup(app: Sphinx):
    app.connect("config-inited", init_db)
    app.connect("source-read", clear_diagnostics)

    # TODO: Support for Sphinx v7+
    # app.connect("include-read")

    # sync_diagnostics needs to run after all diagnostics have been emitted.
    # This ensures proper synchronization of data, as some extensions, like
    # sphinx-needs, emit diagnostics during the "build-finished" event.
    # Setting priority to 900 ensures this handler runs after typical operations.
    app.connect("build-finished", sync_diagnostics, priority=900)



from __future__ import annotations

import ast
import logging
import pathlib
import sys
import typing

from sphinx.application import Sphinx as _Sphinx
from sphinx.errors import ThemeError
from sphinx.util import console
from sphinx.util import logging as sphinx_logging_module
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE

from . import types
from .database import Database
from .log import DiagnosticFilter

if typing.TYPE_CHECKING:
    from typing import IO
    from typing import Any
    from typing import Literal

    from docutils.nodes import Element
    from docutils.parsers.rst import Directive

    DirectiveDefinition = tuple[
        str, type[Directive], list[types.Directive.ArgumentProvider]
    ]
    RoleDefinition = tuple[str, Any, list[types.Role.TargetProvider]]


sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)
logger = sphinx_logger.getChild("esbonio")
sphinx_log_setup = sphinx_logging_module.setup


def setup_logging(app: Sphinx, status: IO, warning: IO):
    # Run the usual setup
    sphinx_log_setup(app, status, warning)

    # Attach our diagnostic filter to the warning handler.
    for handler in sphinx_logger.handlers:
        if handler.level == logging.WARNING:
            handler.addFilter(app.esbonio.log)


class Esbonio:
    """Esbonio specific functionality."""

    db: Database

    log: DiagnosticFilter

    def __init__(self, dbpath: pathlib.Path, app: _Sphinx):
        self.app: _Sphinx = app
        self.db = Database(dbpath)
        self.log = DiagnosticFilter(app)

        self._roles: list[RoleDefinition] = []
        """Roles captured during Sphinx startup."""

        self._directives: list[DirectiveDefinition] = []
        """Directives captured during Sphinx startup."""

        self._config_ast: ast.Module | Literal[False] | None = None
        """The parsed AST of the user's conf.py file.
        If ``False``, we already tried parsing the module and were unable to."""

        self.diagnostics: dict[types.Uri, set[types.Diagnostic]] = {}
        """Recorded diagnostics."""

    @property
    def config_uri(self) -> types.Uri:
        return types.Uri.for_file(pathlib.Path(self.app.confdir, "conf.py"))

    @property
    def config_ast(self) -> ast.Module | None:
        """Return the AST for the user's conf.py (if possible)"""

        if self._config_ast is not None:
            return self._config_ast or None  # convert ``False`` to ``None``

        try:
            conf_py = pathlib.Path(self.app.confdir, "conf.py")
            self._config_ast = ast.parse(source=conf_py.read_text())
            return self._config_ast
        except Exception as exc:
            logger.debug("Unable to parse user's conf.py: %s", exc)
            self._config_ast = False

            return None

    def add_directive(
        self,
        name: str,
        directive: type[Directive],
        argument_providers: list[types.Directive.ArgumentProvider] | None = None,
    ):
        """Register a directive with esbonio.

        Parameters
        ----------
        name
           The name of the directive, as the user would type in a document

        directive
           The directive's implementation

        argument_providers
           A list of argument providers for the role
        """
        self._directives.append((name, directive, argument_providers or []))

    @staticmethod
    def create_directive_argument_provider(
        name: str, **kwargs
    ) -> types.Directive.ArgumentProvider:
        """Create a new directive argument provider

        Parameters
        ----------
        name
           The name of the provider

        kwargs
           Additional arguments to pass to the provider instance

        Returns
        -------
        types.Directive.ArgumentProvider
           The target provider
        """
        return types.Directive.ArgumentProvider(name, kwargs)

    def add_role(
        self,
        name: str,
        role: Any,
        target_providers: list[types.Role.TargetProvider] | None = None,
    ):
        """Register a role with esbonio.

        Parameters
        ----------
        name
           The name of the role, as the user would type in a document

        role
           The role's implementation

        target_providers
           A list of target providers for the role
        """
        self._roles.append((name, role, target_providers or []))

    @staticmethod
    def create_role_target_provider(name: str, **kwargs) -> types.Role.TargetProvider:
        """Create a new role target provider

        Parameters
        ----------
        name
           The name of the provider

        kwargs
           Additional arguments to pass to the provider instance

        Returns
        -------
        types.Role.TargetProvider
           The target provider
        """
        return types.Role.TargetProvider(name, kwargs)


class Sphinx(_Sphinx):
    """An extended sphinx application that integrates with esbonio."""

    esbonio: Esbonio

    def __init__(self, *args, **kwargs):
        # Disable color codes
        console.nocolor()

        # Add in esbonio specific functionality
        self.esbonio = Esbonio(
            dbpath=pathlib.Path(kwargs["outdir"], "esbonio.db").resolve(),
            app=self,
        )

        # Override sphinx's usual logging setup function
        sphinx_logging_module.setup = setup_logging  # type: ignore

        # `try_run_init` may call `__init__` more than once, this could lead to spamming
        # the user with warning messages, so we will suppress these messages if the
        # retry counter has been set.
        self._esbonio_retry_count = 0
        try_run_init(self, super().__init__, *args, **kwargs)

    def add_role(self, name: str, role: Any, override: bool = False):
        super().add_role(name, role, override or self._esbonio_retry_count > 0)
        self.esbonio.add_role(name, role)

    def add_directive(self, name: str, cls: type[Directive], override: bool = False):
        super().add_directive(name, cls, override or self._esbonio_retry_count > 0)
        self.esbonio.add_directive(name, cls)

    def add_node(self, node: type[Element], override: bool = False, **kwargs):
        super().add_node(node, override or self._esbonio_retry_count > 0, **kwargs)

    def setup_extension(self, extname: str):
        """Override Sphinx's implementation of `setup_extension`

        This implementation
        - Will suppress errors caused by missing extensions
        - Attempt to report errors where possible, as diagnostics
        """
        try:
            super().setup_extension(extname)
        except Exception as exc:
            # Attempt to produce useful diagnostics.
            self._report_missing_extension(extname, exc)

    def _report_missing_extension(self, extname: str, exc: Exception):
        """Check to see if the given exception corresponds to a missing extension.

        If so, attempt to produce a diagnostic to highlight this to the user.

        Parameters
        ----------
        extname
           The name of the extension that caused the exception

        exc
           The exception instance
        """

        if (config := self.esbonio.config_ast) is None:
            return

        # Now attempt to find the soure location of the extenison.
        if (range_ := find_extension_declaration(config, extname)) is None:
            logger.debug("Unable to locate declaration of extension: %r", extname)
            return

        diagnostic = types.Diagnostic(
            range=range_, message=f"{exc}", severity=types.DiagnosticSeverity.Error
        )

        uri = self.esbonio.config_uri
        logger.debug("Adding diagnostic %s: %s", uri, diagnostic)
        self.esbonio.diagnostics.setdefault(uri, set()).add(diagnostic)


def try_run_init(app: Sphinx, init_fn, *args, **kwargs):
    """Try and run Sphinx's ``__init__`` function.

    There are occasions where Sphinx will try and throw an error that is recoverable
    e.g. a missing theme. In these situations we want to suppress the error, record a
    diagnostic, and try again - which is what this function will do.

    Some errors however, are not recoverable in which case we will allow the error to
    proceed as normal.

    Parameters
    ----------
    app
       The application instance we are trying to initialize

    init_fn
       The application's `__init__` method, as returned by ``super().__init__``

    args
       Positional arguments to ``__init__``

    retries
       Max number of retries, a fallback in case we end up creating infinite recursion

    kwargs
       Keyword arguments to ``__init__``
    """

    if app._esbonio_retry_count >= 100:
        raise RuntimeError("Unable to initialize Sphinx: max retries exceeded")

    try:
        init_fn(*args, **kwargs)
    except ThemeError as exc:
        # Fallback to the default theme.
        print(f"ThemeError: {exc}", file=sys.stderr)  # noqa: T201

        fallback_theme = "alabaster"
        kwargs.setdefault("confoverrides", {})["html_theme"] = fallback_theme
        kwargs["confoverrides"]["html_theme_options"] = {}

        app._esbonio_retry_count += 1
        report_theme_error(app, exc)

        print(f"Retrying with html_theme = {fallback_theme!r}\n", file=sys.stderr)  # noqa: T201
        try_run_init(app, init_fn, *args, **kwargs)
    except Exception:
        logger.exception("Unable to initialize Sphinx")
        raise


def report_theme_error(app: Sphinx, exc: ThemeError):
    """Attempt to convert the given theme error into a useful diagnostic.

    Parameters
    ----------
    app
       The Sphinx object being initialized

    exc
       The error instance
    """

    if (config := app.esbonio.config_ast) is None:
        return

    if (range_ := find_html_theme_declaration(config)) is None:
        return

    diagnostic = types.Diagnostic(
        range=range_,
        message=f"{exc}",
        severity=types.DiagnosticSeverity.Error,
    )

    uri = app.esbonio.config_uri
    logger.debug("Adding diagnostic %s: %s", uri, diagnostic)
    app.esbonio.diagnostics.setdefault(uri, set()).add(diagnostic)


def find_html_theme_declaration(mod: ast.Module) -> types.Range | None:
    """Attempt to find the location in the user's conf.py file where the ``html_theme``
    was declared."""

    for node in mod.body:
        if not isinstance(node, ast.Assign):
            continue

        if len(targets := node.targets) != 1:
            continue

        if not isinstance(name := targets[0], ast.Name):
            continue

        if name.id == "html_theme":
            break

    else:
        # Nothing found, abort
        logger.debug("Unable to find 'html_theme' node")
        return None

    return ast_node_to_range(node)


def find_extension_declaration(mod: ast.Module, extname: str) -> types.Range | None:
    """Attempt to find the location in the user's conf.py file where the given
    ``extname`` was declared.

    This function will never be perfect (conf.py is after all, turing complete!).
    However, it *should* be possible to write something that can handle most cases.
    """

    # First try and locate the node corresponding to `extensions = [ ... ]`
    for node in mod.body:
        if not isinstance(node, ast.Assign):
            continue

        if len(targets := node.targets) != 1:
            continue

        if not isinstance(name := targets[0], ast.Name):
            continue

        if name.id == "extensions":
            break

    else:
        # Nothing found, abort
        logger.debug("Unable to find 'extensions' node")
        return None

    # Now try to find the node corresponding to `'extname'`
    if not isinstance(extlist := node.value, ast.List):
        return None

    for element in extlist.elts:
        if not isinstance(element, ast.Constant):
            continue

        if element.value == extname:
            break
    else:
        # Nothing found, abort
        logger.debug("Unable to find node for extension %r", extname)
        return None

    return ast_node_to_range(element)


def ast_node_to_range(node: ast.stmt | ast.expr) -> types.Range:
    """Convert the given ast node to a range."""

    # Finally, try and extract the source location.
    start_line = node.lineno - 1
    start_char = node.col_offset

    if (end_line := (node.end_lineno or 0) - 1) < 0:
        end_line = start_line + 1
        end_char: int | None = 0

    elif (end_char := node.end_col_offset) is None:
        end_line += 1
        end_char = 0

    return types.Range(
        start=types.Position(line=start_line, character=start_char),
        end=types.Position(line=end_line, character=end_char or 0),
    )



import logging
import os
import pathlib
import sys

import pytest
from lsprotocol.types import WorkspaceFolder
from pygls import IS_WIN
from pygls.workspace import Workspace

from esbonio.server import Uri
from esbonio.server.features.sphinx_manager import SphinxConfig
from esbonio.server.features.sphinx_manager import SubProcess
from esbonio.server.features.sphinx_manager.config import get_module_path

logger = logging.getLogger(__name__)


# Default values
PYTHON_CMD = ["/bin/python"]
BUILD_CMD = ["sphinx-build", "-M", "html", "src", "dest"]
PYPATH = [pathlib.Path("/path/to/site-packages/esbonio")]
ENV = {
    "PYTHONUNBUFFERED": "1",
    "PYTHONPATH": str(get_module_path("esbonio.sphinx_agent")),
}
CWD = r"c:\path\to\workspace" if IS_WIN else "/path/to/workspace"


# The value of FALLBACK_ENV must actually exist somewhere on the filesystem
# But for the tests here, the actual location doesn't really matter
FALLBACK_ENV = str(pathlib.Path(__file__).parent)


def mk_uri(path: str) -> str:
    return str(Uri.for_file(path))


@pytest.mark.parametrize(
    "uri, workspace, config, expected",
    [
        (  # If everything is specified, resolve should be a no-op
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
        ),
        (  # If no cwd is given, and there is no available workspace root the config
            # should be considered invalid
            "file:///path/to/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            None,
        ),
        (  # If the workspace is empty, we should still be able to progress as long as
            # the user provides a cwd
            "file:///path/to/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
        ),
        pytest.param(  # If only a ``root_uri`` is given use that.
            "file:///path/to/workspace/file.rst",
            Workspace(mk_uri(CWD)),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # If only a ``root_uri`` is given use that.
            "file:///c:/path/to/workspace/file.rst",
            Workspace(mk_uri(CWD)),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        (  # Assuming that the requested uri resides within it.
            "file:///path/to/other/workspace/file.rst",
            Workspace(mk_uri(CWD)),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            None,
        ),
        pytest.param(  # Otherwise, prefer workspace_folders.
            "file:///path/to/workspace/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[WorkspaceFolder(mk_uri(CWD), "workspace")],
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(  # Otherwise, prefer workspace_folders.
            "file:///c:/path/to/workspace/file.rst",
            Workspace(
                "file:///c:/path/to",
                workspace_folders=[WorkspaceFolder(mk_uri(CWD), "workspace")],
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD, cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        (  # Handle multi-root scenarios.
            "file:///path/to/workspace-b/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[
                    WorkspaceFolder("file:///path/to/workspace-a", "workspace-a"),
                    WorkspaceFolder("file:///path/to/workspace-b", "workspace-b"),
                ],
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD,
                    cwd=os.path.join(".", "path", "to", "workspace-b")[1:],
                    env=ENV,
                ),
                build_command=BUILD_CMD,
            ),
        ),
        (  # Again, make sure the requested uri resides within the workspace.
            "file:///path/for/workspace-c/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[
                    WorkspaceFolder("file:///path/to/workspace-a", "workspace-a"),
                    WorkspaceFolder("file:///path/to/workspace-b", "workspace-b"),
                ],
            ),
            SphinxConfig(
                python_command=SubProcess(command=PYTHON_CMD),
                build_command=BUILD_CMD,
            ),
            None,
        ),
        (  # If no python command provided, fallback to the server's environment
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=[], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(command=[sys.executable], cwd=CWD, env=ENV),
                build_command=BUILD_CMD,
            ),
        ),
        (  # Allow the user to specify extra envrionment variables
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD, cwd=CWD, env={"MY_VAR": "some-value"}
                ),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD,
                    cwd=CWD,
                    env={
                        **ENV,
                        "MY_VAR": "some-value",
                    },
                ),
                build_command=BUILD_CMD,
            ),
        ),
        (  # Or override existing variables in the envrionment
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD, cwd=CWD, env={"HOME": "/home/not-really-a-user"}
                ),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD,
                    cwd=CWD,
                    env={
                        **ENV,
                        "HOME": "/home/not-really-a-user",
                    },
                ),
                build_command=BUILD_CMD,
            ),
        ),
        (  # But if the user sets their own PYTHONPATH, be sure to include it alongside our own
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD, cwd=CWD, env={"PYTHONPATH": "/extra/py/path"}
                ),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=PYTHON_CMD,
                    cwd=CWD,
                    env={
                        **ENV,
                        "PYTHONPATH": f"{ENV['PYTHONPATH']}{os.pathsep}/extra/py/path",
                    },
                ),
                build_command=BUILD_CMD,
            ),
        ),
        pytest.param(
            # When not on Windows `${venv:/path/to/env}` should expand to
            # `/path/to/env/bin/python`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:/path/to/env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["/path/to/env/bin/python"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            # When on Windows `${venv:c:/path/to/env}` should expand to
            # `c:/path/to/env/Scripts/python.exe`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:c:/path/to/env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["c:\\path\\to\\env\\Scripts\\python.exe"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            # When on Windows `${venv:c:\\path\\to\\env}` should expand to
            # `c:/path/to/env/Scripts/python.exe`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(
                    command=["${venv:c:\\path\\to\\env}"], cwd=CWD
                ),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["c:\\path\\to\\env\\Scripts\\python.exe"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            # When not on Windows `${venv:env}` should expand to
            # `${cwd}/env/bin/python`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["/path/to/workspace/env/bin/python"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            # When on Windows `${venv:env}` should expand to
            # `c:/path/to/workspace/env/Scripts/python.exe`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["c:\\path\\to\\workspace\\env\\Scripts\\python.exe"],
                    cwd=CWD,
                    env=ENV,
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            # When not on Windows `${venv:../env}` should expand to
            # `${cwd}/../env/bin/python`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:../env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["/path/to/env/bin/python"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(IS_WIN, reason="windows"),
        ),
        pytest.param(
            # When on Windows `${venv:../env}` should expand to
            # `c:/path/to/workspace/Scripts/python.exe`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:../env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["c:\\path\\to\\env\\Scripts\\python.exe"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
        pytest.param(
            # When on Windows `${venv:..\\env}` should expand to
            # `c:/path/to/workspace/Scripts/python.exe`
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=SubProcess(command=["${venv:..\\env}"], cwd=CWD),
                build_command=BUILD_CMD,
            ),
            SphinxConfig(
                python_command=SubProcess(
                    command=["c:\\path\\to\\env\\Scripts\\python.exe"], cwd=CWD, env=ENV
                ),
                build_command=BUILD_CMD,
            ),
            marks=pytest.mark.skipif(not IS_WIN, reason="windows only"),
        ),
    ],
)
def test_resolve(
    uri: str,
    workspace: Workspace,
    config: SphinxConfig,
    expected: SphinxConfig | None,
):
    """Ensure that we can resolve a user's configuration correctly.

    Parameters
    ----------
    uri
       The uri the config should be resolved relative to

    workspace
       The workspace in which to resolve the configuration

    config
       The base configuration to resolve

    expected
       The expected outcome
    """
    actual = config.resolve(Uri.parse(uri), workspace, logger)

    if expected is None:
        assert actual is None
        return

    assert actual is not None
    assert actual.enable_dev_tools == expected.enable_dev_tools
    assert actual.build_command == expected.build_command
    assert actual.config_overrides == expected.config_overrides

    assert actual.python_command.command == expected.python_command.command
    assert actual.python_command.cwd == expected.python_command.cwd

    # Since we pass through the current environment, check each expected key individually
    for varname in expected.python_command.env:
        assert (
            actual.python_command.env[varname] == expected.python_command.env[varname]
        )



from __future__ import annotations

import asyncio
import logging
import socket
import threading
import typing

from lsprotocol import types
from pygls.io_ import run_websocket
from pygls.protocol import JsonRPCProtocol
from pygls.protocol import default_converter
from pygls.server import JsonRPCServer
from websockets.asyncio.server import serve

from esbonio import server

if typing.TYPE_CHECKING:
    from websockets.asyncio.server import Server as WebSocketServer
    from websockets.asyncio.server import ServerConnection as WebSocketConnection

    from .config import PreviewConfig


class WebviewServer(JsonRPCServer):
    """The webview server controlls the webpage hosting the preview.

    Used to implement automatic reloads and features like sync scrolling.
    """

    protocol: JsonRPCProtocol

    def __init__(self, logger: logging.Logger, config: PreviewConfig, *args, **kwargs):
        super().__init__(JsonRPCProtocol, default_converter, *args, **kwargs)

        self.config = config
        self.logger = logger.getChild("WebviewServer")

        self._connected = False
        self._ws_server: WebSocketServer | None = None

        self._startup_task: asyncio.Task | None = None
        """The task that resolves once startup is complete."""

        self._server_task: asyncio.Task | None = None
        """The task hosting the server itself."""

        self._editor_in_control: asyncio.Task | None = None
        """If set, the editor is in control and the view should not emit scroll events"""

        self._view_in_control: asyncio.Task | None = None
        """If set, the view is in control and the editor should not emit scroll events"""

        self._current_uri: str | None = None
        """If set, indicates the current uri the editor and view are scrolling."""

    def __await__(self):
        """Makes the server await-able"""
        if self._startup_task is None:
            self._startup_task = asyncio.create_task(self.start())

        return self._startup_task.__await__()

    @property
    def port(self):
        if self._ws_server is None:
            return None

        sock = list(self._ws_server.sockets)[0]
        return sock.getsockname()[1]

    @property
    def connected(self) -> bool:
        """Indicates when we have an active connection to the client."""
        return self._connected

    def reload(self):
        """Reload the current view."""
        if self.connected:
            self.protocol.notify("view/reload", {})

    def scroll(self, uri: str, line: int):
        """Called by the editor to scroll the current webview."""
        if not self.connected or self._view_in_control:
            return

        if self.config.synchronize_scroll not in {"bothWays", "editorWithPreview"}:
            return

        # If the editor is already in control, reset the cooldown
        if self._editor_in_control:
            self._editor_in_control.cancel()

        self._current_uri = uri
        self._editor_in_control = asyncio.create_task(self.cooldown("editor"))
        self.protocol.notify("view/scroll", {"uri": uri, "line": line})

    async def cooldown(self, name: str):
        """Create a cooldown."""
        await asyncio.sleep(1)

        # Unset the cooldown
        self.logger.debug("%s cooldown ended", name)
        setattr(self, f"_{name}_in_control", None)

    async def start(self):
        """Start the server and wrap the server coroutine in a task."""
        self._server_task = asyncio.create_task(
            self._start_ws(self.config.bind, self.config.ws_port)
        )

        # HACK: we need to yield control to the event loop to give the ws_server time to
        #       spin up and allocate a port number.
        await asyncio.sleep(1)
        return self

    def stop(self):
        """Stop the server."""
        self.logger.debug("Shutting down preview WebSocket server")

        if self._server_task is not None:
            self._server_task.cancel()

    async def _start_ws(self, host: str, port: int) -> None:
        """Actually, start the server."""

        async def connection(websocket: WebSocketConnection):
            self._connected = True
            self.logger.debug("Connected")

            await run_websocket(
                stop_event=threading.Event(),
                websocket=websocket,
                protocol=self.protocol,
                logger=self.logger,
                error_handler=self.report_server_error,
            )

            self.logger.debug("Connection lost")
            self._connected = False

        self._ws_server = await serve(
            connection,
            host,
            port,
            # logger=self.logger.getChild("ws"),
            family=socket.AF_INET,  # Use IPv4 only.
        )
        async with self._ws_server:
            await self._ws_server.serve_forever()


def make_ws_server(
    esbonio: server.EsbonioLanguageServer, config: PreviewConfig
) -> WebviewServer:
    server = WebviewServer(esbonio.logger, config)

    @server.feature("editor/scroll")
    def on_scroll(ls: WebviewServer, params):
        """Called by the webview to scroll the editor."""
        if not server.connected or server._editor_in_control:
            return

        if ls.config.synchronize_scroll not in {"bothWays", "previewWithEditor"}:
            return

        # If the view is already in control, reset the cooldown.
        if server._view_in_control:
            server._view_in_control.cancel()

        server._view_in_control = asyncio.create_task(server.cooldown("view"))

        esbonio.window_show_document(
            types.ShowDocumentParams(
                uri=params.uri,
                external=False,
                selection=types.Range(
                    start=types.Position(line=params.line - 1, character=0),
                    end=types.Position(line=params.line, character=0),
                ),
            )
        )

    return server



import pytest

from esbonio.sphinx_agent.types import MYST_DIRECTIVE
from esbonio.sphinx_agent.types import MYST_ROLE
from esbonio.sphinx_agent.types import RST_DEFAULT_ROLE
from esbonio.sphinx_agent.types import RST_DIRECTIVE
from esbonio.sphinx_agent.types import RST_DIRECTIVE_OPTION
from esbonio.sphinx_agent.types import RST_ROLE


@pytest.mark.parametrize(
    "string, expected",
    [
        ("`", None),
        ("``", None),
        ("```", {"directive": "```"}),
        ("````", {"directive": "````"}),
        ("```{d", {"directive": "```{d", "name": "d"}),
        ("```python", {"directive": "```python", "name": "python"}),
        ("```{image}", {"name": "image", "directive": "```{image}"}),
        (
            "```{image}  filename.png",
            {"name": "image", "directive": "```{image}", "argument": "filename.png"},
        ),
        (
            "```{image}  filename.png  \n",
            {"name": "image", "directive": "```{image}", "argument": "filename.png"},
        ),
    ],
)
def test_myst_directive_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse directives
    works as expected.

    As with most test cases, this one is parameterized with the following arguments::

       (".. figure::", {"name": "figure"})

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result
    will **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situaions where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = MYST_DIRECTIVE.match(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value, name


@pytest.mark.parametrize(
    "string, expected",
    [
        ("`{", None),
        ("d{", None),
        ("{", {"role": "{"}),
        ("{ref", {"name": "ref", "role": "{ref"}),
        # The pattern should still work if the user adds a role in the middle of a line
        ("{ref for more details", {"name": "ref", "role": "{ref"}),
        ("{ref} for more details", {"name": "ref", "role": "{ref}"}),
        ("{code-block", {"name": "code-block", "role": "{code-block"}),
        ("{c:func}", {"name": "c:func", "role": "{c:func}"}),
        ("{cpp:func}", {"name": "cpp:func", "role": "{cpp:func}"}),
        (
            "{external:cpp:func}",
            {"name": "external:cpp:func", "role": "{external:cpp:func}"},
        ),
        (
            "{external+python:cpp:func}",
            {"name": "external+python:cpp:func", "role": "{external+python:cpp:func}"},
        ),
        ("{ref}`", {"name": "ref", "role": "{ref}", "target": "`"}),
        (
            "{code-block}`",
            {"name": "code-block", "role": "{code-block}", "target": "`"},
        ),
        (
            "{c:func}`",
            {"name": "c:func", "role": "{c:func}", "target": "`"},
        ),
        (
            "{ref}`some_label",
            {
                "name": "ref",
                "role": "{ref}",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            "{ref}`!some_label",
            {
                "name": "ref",
                "role": "{ref}",
                "label": "some_label",
                "target": "`!some_label",
                "modifier": "!",
            },
        ),
        (
            "{ref}`~some_label",
            {
                "name": "ref",
                "role": "{ref}",
                "label": "some_label",
                "target": "`~some_label",
                "modifier": "~",
            },
        ),
        (
            "{code-block}`some_label",
            {
                "name": "code-block",
                "role": "{code-block}",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            "{c:func}`some_label",
            {
                "name": "c:func",
                "role": "{c:func}",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            "{ref}`some_label`",
            {
                "name": "ref",
                "role": "{ref}",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            "{code-block}`some_label`",
            {
                "name": "code-block",
                "role": "{code-block}",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            "{c:func}`some_label`",
            {
                "name": "c:func",
                "role": "{c:func}",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            "{ref}`see more <",
            {
                "name": "ref",
                "role": "{ref}",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            "{code-block}`see more <",
            {
                "name": "code-block",
                "role": "{code-block}",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            "{c:func}`see more <",
            {
                "name": "c:func",
                "role": "{c:func}",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            "{ref}`see more <some_label",
            {
                "name": "ref",
                "role": "{ref}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            "{code-block}`see more <some_label",
            {
                "name": "code-block",
                "role": "{code-block}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            "{ref}`see more <!some_label",
            {
                "name": "ref",
                "role": "{ref}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <!some_label",
                "modifier": "!",
            },
        ),
        (
            "{code-block}`see more <~some_label",
            {
                "name": "code-block",
                "role": "{code-block}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <~some_label",
                "modifier": "~",
            },
        ),
        (
            "{c:func}`see more <some_label",
            {
                "name": "c:func",
                "role": "{c:func}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            "{ref}`see more <some_label>",
            {
                "name": "ref",
                "role": "{ref}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            "{code-block}`see more <some_label>",
            {
                "name": "code-block",
                "role": "{code-block}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            "{c:func}`see more <some_label>",
            {
                "name": "c:func",
                "role": "{c:func}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            "{ref}`see more <some_label>`",
            {
                "name": "ref",
                "role": "{ref}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
        (
            "{code-block}`see more <some_label>`",
            {
                "name": "code-block",
                "role": "{code-block}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
        (
            "{c:func}`see more <some_label>`",
            {
                "name": "c:func",
                "role": "{c:func}",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
    ],
)
def test_myst_role_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse roles works as
    expected.

    As a general rule, it's better to write tests at the LSP protocol level as that
    decouples the test cases from the implementation. However, roles and the
    corresponding regular expression are complex enough to warrant a test case on its
    own.

    As with most test cases, this one is parameterized with the following arguments::

        ((":ref:", {"name": "ref"}),)
        (".. directive::", None)

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result will
    **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situations where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = MYST_ROLE.search(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value


@pytest.mark.parametrize(
    "string, expected",
    [
        (".", None),
        ("..", {"directive": ".."}),
        (".. d", {"directive": ".. d"}),
        (".. image::", {"name": "image", "directive": ".. image::"}),
        (".. c:", {"name": "c:", "directive": ".. c:"}),
        (".. |", {"directive": ".. |", "substitution": "|"}),
        (
            ".. |e",
            {"directive": ".. |e", "substitution": "|e", "substitution_text": "e"},
        ),
        (
            ".. |example",
            {
                "directive": ".. |example",
                "substitution": "|example",
                "substitution_text": "example",
            },
        ),
        (
            ".. |example|",
            {
                "directive": ".. |example|",
                "substitution": "|example|",
                "substitution_text": "example",
            },
        ),
        (
            ".. |example| replace::",
            {
                "directive": ".. |example| replace::",
                "substitution": "|example|",
                "substitution_text": "example",
                "name": "replace",
            },
        ),
        (
            ".. |example| replace:: some text here",
            {
                "directive": ".. |example| replace::",
                "substitution": "|example|",
                "substitution_text": "example",
                "name": "replace",
                "argument": "some text here",
            },
        ),
        (
            ".. c:function::",
            {"name": "c:function", "directive": ".. c:function::"},
        ),
        (
            ".. image:: filename.png",
            {"name": "image", "argument": "filename.png", "directive": ".. image::"},
        ),
        (
            ".. image:: filename.png  ",
            {"name": "image", "argument": "filename.png", "directive": ".. image::"},
        ),
        (
            ".. image:: filename.png\r",
            {"name": "image", "argument": "filename.png", "directive": ".. image::"},
        ),
        (
            ".. cpp:function:: malloc",
            {
                "name": "cpp:function",
                "argument": "malloc",
                "directive": ".. cpp:function::",
            },
        ),
        (
            "   .. image:: filename.png",
            {"name": "image", "argument": "filename.png", "directive": ".. image::"},
        ),
        (
            "   .. cpp:function:: malloc",
            {
                "name": "cpp:function",
                "argument": "malloc",
                "directive": ".. cpp:function::",
            },
        ),
        (
            ".. rst:directive:option::",
            {
                "name": "rst:directive:option",
                "directive": ".. rst:directive:option::",
            },
        ),
        (
            "   .. rst:directive:option::",
            {
                "name": "rst:directive:option",
                "directive": ".. rst:directive:option::",
            },
        ),
        (
            "   .. rst:directive:option:: height",
            {
                "name": "rst:directive:option",
                "directive": ".. rst:directive:option::",
                "argument": "height",
            },
        ),
    ],
)
def test_rst_directive_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse directives
    works as expected.

    As with most test cases, this one is parameterized with the following arguments::

       (".. figure::", {"name": "figure"})

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result
    will **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situaions where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = RST_DIRECTIVE.match(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value, name


@pytest.mark.parametrize(
    "string, expected",
    [
        (":", None),
        (":align", None),
        (":align:", None),
        (":align: center", None),
        ("   :", {"indent": "   ", "option": ":"}),
        ("   :align", {"indent": "   ", "option": ":align", "name": "align"}),
        ("   :align:", {"indent": "   ", "option": ":align:", "name": "align"}),
        (
            "   :align: center",
            {"indent": "   ", "option": ":align:", "name": "align", "value": "center"},
        ),
    ],
)
def test_directive_option_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse directive
    options works as expected.

    As with most test cases, this one is parameterized with the following arguments::

       ("   :align:", {"name": "align"})

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result
    will **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situaions where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = RST_DIRECTIVE_OPTION.match(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value


@pytest.mark.parametrize(
    "string, expected",
    [
        ("::", None),
        (":", {"role": ":"}),
        (":ref", {"name": "ref", "role": ":ref"}),
        # The pattern should still work if the user adds a role in the middle of a line
        (":ref for more details", {"name": "ref", "role": ":ref"}),
        (":ref: for more details", {"name": "ref", "role": ":ref:"}),
        (":code-block", {"name": "code-block", "role": ":code-block"}),
        (":c:func:", {"name": "c:func", "role": ":c:func:"}),
        (":cpp:func:", {"name": "cpp:func", "role": ":cpp:func:"}),
        (
            ":external:cpp:func:",
            {"name": "external:cpp:func", "role": ":external:cpp:func:"},
        ),
        (
            ":external+python:cpp:func:",
            {"name": "external+python:cpp:func", "role": ":external+python:cpp:func:"},
        ),
        (":ref:`", {"name": "ref", "role": ":ref:", "target": "`"}),
        (
            ":code-block:`",
            {"name": "code-block", "role": ":code-block:", "target": "`"},
        ),
        (
            ":c:func:`",
            {"name": "c:func", "role": ":c:func:", "target": "`"},
        ),
        (
            ":ref:`some_label",
            {
                "name": "ref",
                "role": ":ref:",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            ":ref:`!some_label",
            {
                "name": "ref",
                "role": ":ref:",
                "label": "some_label",
                "target": "`!some_label",
                "modifier": "!",
            },
        ),
        (
            ":ref:`~some_label",
            {
                "name": "ref",
                "role": ":ref:",
                "label": "some_label",
                "target": "`~some_label",
                "modifier": "~",
            },
        ),
        (
            ":code-block:`some_label",
            {
                "name": "code-block",
                "role": ":code-block:",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            ":c:func:`some_label",
            {
                "name": "c:func",
                "role": ":c:func:",
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            ":ref:`some_label`",
            {
                "name": "ref",
                "role": ":ref:",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            ":code-block:`some_label`",
            {
                "name": "code-block",
                "role": ":code-block:",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            ":c:func:`some_label`",
            {
                "name": "c:func",
                "role": ":c:func:",
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            ":ref:`see more <",
            {
                "name": "ref",
                "role": ":ref:",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            ":code-block:`see more <",
            {
                "name": "code-block",
                "role": ":code-block:",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            ":c:func:`see more <",
            {
                "name": "c:func",
                "role": ":c:func:",
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            ":ref:`see more <some_label",
            {
                "name": "ref",
                "role": ":ref:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            ":code-block:`see more <some_label",
            {
                "name": "code-block",
                "role": ":code-block:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            ":ref:`see more <!some_label",
            {
                "name": "ref",
                "role": ":ref:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <!some_label",
                "modifier": "!",
            },
        ),
        (
            ":code-block:`see more <~some_label",
            {
                "name": "code-block",
                "role": ":code-block:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <~some_label",
                "modifier": "~",
            },
        ),
        (
            ":c:func:`see more <some_label",
            {
                "name": "c:func",
                "role": ":c:func:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            ":ref:`see more <some_label>",
            {
                "name": "ref",
                "role": ":ref:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            ":code-block:`see more <some_label>",
            {
                "name": "code-block",
                "role": ":code-block:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            ":c:func:`see more <some_label>",
            {
                "name": "c:func",
                "role": ":c:func:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>",
            },
        ),
        (
            ":ref:`see more <some_label>`",
            {
                "name": "ref",
                "role": ":ref:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
        (
            ":code-block:`see more <some_label>`",
            {
                "name": "code-block",
                "role": ":code-block:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
        (
            ":c:func:`see more <some_label>`",
            {
                "name": "c:func",
                "role": ":c:func:",
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
    ],
)
def test_rst_role_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse roles works as
    expected.

    As a general rule, it's better to write tests at the LSP protocol level as that
    decouples the test cases from the implementation. However, roles and the
    corresponding regular expression are complex enough to warrant a test case on its
    own.

    As with most test cases, this one is parameterized with the following arguments::

        ((":ref:", {"name": "ref"}),)
        (".. directive::", None)

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result will
    **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situations where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = RST_ROLE.search(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value


@pytest.mark.parametrize(
    "string, expected",
    [
        (
            "`",
            {"target": "`"},
        ),
        (
            ":ref:`",
            None,
        ),
        (
            ":ref:``",
            None,
        ),
        (
            ":py:func:`",
            None,
        ),
        (
            ":py:func:``",
            None,
        ),
        (
            "`!some_label",
            {
                "label": "some_label",
                "target": "`!some_label",
                "modifier": "!",
            },
        ),
        (
            "`~some_label",
            {
                "label": "some_label",
                "target": "`~some_label",
                "modifier": "~",
            },
        ),
        (
            "`some_label",
            {
                "label": "some_label",
                "target": "`some_label",
            },
        ),
        (
            "`some_label`",
            {
                "label": "some_label",
                "target": "`some_label`",
            },
        ),
        (
            "`see more <",
            {
                "alias": "see more ",
                "target": "`see more <",
            },
        ),
        (
            "`see more <!some_label",
            {
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <!some_label",
                "modifier": "!",
            },
        ),
        (
            "`see more <~some_label",
            {
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <~some_label",
                "modifier": "~",
            },
        ),
        (
            "`see more <some_label",
            {
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label",
            },
        ),
        (
            "`see more <some_label>`",
            {
                "alias": "see more ",
                "label": "some_label",
                "target": "`see more <some_label>`",
            },
        ),
    ],
)
def test_default_role_regex(string, expected):
    """Ensure that the regular expression we use to detect and parse default roles works
    as expected.

    As a general rule, it's better to write tests at the LSP protocol level as that
    decouples the test cases from the implementation. However, roles and the
    corresponding regular expression are complex enough to warrant a test case on its
    own.

    As with most test cases, this one is parameterized with the following arguments::

        ((":ref:", {"name": "ref"}),)
        (".. directive::", None)

    The first argument is the string to test the pattern against, the second a
    dictionary containing the groups we expect to see in the resulting match object.
    Groups that appear in the resulting match object but not in the expected result will
    **not** fail the test.

    To test situations where the pattern should **not** match the input, pass ``None``
    as the second argument.

    To test situations where the pattern should match, but we don't expect to see any
    groups pass an empty dictionary as the second argument.
    """

    match = RST_DEFAULT_ROLE.search(string)

    if expected is None:
        assert match is None
    else:
        assert match is not None

        for name, value in expected.items():
            assert match.groupdict().get(name, None) == value



from __future__ import annotations

import re

import pytest
from lsprotocol import types
from pygls.workspace import TextDocument

from esbonio import server


@pytest.mark.parametrize(
    "language, languages, expected",
    [
        ("rst", set(), True),
        ("markdown", set(), True),
        ("rst", {"markdown"}, False),
        ("rst", {"rst", "python"}, True),
    ],
)
def test_completion_trigger_languages(
    language: str, languages: set[str], expected: bool
):
    """Ensure that ``CompletionTrigger`` responds to the given language correctly."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=0),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
    )
    document = TextDocument(uri=str(uri), source="some text")

    trigger = server.CompletionTrigger(patterns=[re.compile(".*")], languages=languages)
    result = trigger(uri, params, document, language, types.ClientCapabilities())

    assert (result is not None) == expected


@pytest.mark.parametrize(
    "context, characters, expected",
    [
        # No context, or a non-trigger character request should still trigger
        (None, set(), True),
        (None, {">", "."}, True),
        (types.CompletionContext(types.CompletionTriggerKind.Invoked), set(), True),
        (
            types.CompletionContext(types.CompletionTriggerKind.Invoked),
            {".", "<"},
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerForIncompleteCompletions
            ),
            set(),
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerForIncompleteCompletions
            ),
            {".", "<"},
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerCharacter,
            ),
            set(),
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerCharacter,
            ),
            {".", "<"},
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerCharacter, trigger_character="."
            ),
            set(),
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerCharacter, trigger_character="."
            ),
            {".", ":"},
            True,
        ),
        (
            types.CompletionContext(
                types.CompletionTriggerKind.TriggerCharacter, trigger_character="."
            ),
            {":"},
            False,
        ),
    ],
)
def test_completion_trigger_characters(
    context: types.CompletionContext | None, characters: set[str], expected: bool
):
    """Ensure that ``CompletionTrigger`` responds to the trigger character correctly."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=0),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
        context=context,
    )
    document = TextDocument(uri=str(uri), source="some text")

    trigger = server.CompletionTrigger(
        patterns=[re.compile(".*")], characters=characters
    )
    result = trigger(uri, params, document, "rst", types.ClientCapabilities())

    assert (result is not None) == expected


def test_completion_trigger_pattern_no_match():
    """Ensure that if none of the ``CompletionTrigger`` patterns match, the trigger does
    not fire."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=0),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
    )
    document = TextDocument(uri=str(uri), source="some text")

    trigger = server.CompletionTrigger(patterns=[re.compile("xx"), re.compile("yy")])
    assert trigger(uri, params, document, "rst", types.ClientCapabilities()) is None


def test_completion_trigger_pattern_match_wrong_pos():
    """Ensure that if one of the ``CompletionTrigger`` patterns match, the trigger does
    not fire if the match is outside of the requested location."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=0),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
    )
    document = TextDocument(uri=str(uri), source="some xx text")

    trigger = server.CompletionTrigger(patterns=[re.compile("xx"), re.compile("yy")])
    assert trigger(uri, params, document, "rst", types.ClientCapabilities()) is None


def test_completion_trigger_pattern_match():
    """Ensure that if one of the ``CompletionTrigger`` patterns match, the trigger returns
    the completion context with the correct fields."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=6),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
    )
    document = TextDocument(uri=str(uri), source="some xx text")
    client_capabilities = types.ClientCapabilities(
        text_document=types.TextDocumentClientCapabilities(
            completion=types.CompletionClientCapabilities(
                context_support=True,
            ),
        ),
    )

    trigger = server.CompletionTrigger(patterns=[re.compile("xx"), re.compile("yy")])
    result = trigger(uri, params, document, "rst", client_capabilities)

    assert result is not None
    assert result.uri == uri
    assert result.doc == document
    assert result.position == params.position
    assert result.language == "rst"
    assert result.capabilities == client_capabilities

    assert result.match.group(0) == "xx"


def test_completion_trigger_pattern_first_match():
    """Ensure that if more than one of the ``CompletionTrigger`` patterns match, the
    trigger returns the first match."""

    uri = server.Uri.parse("file:///test.txt")
    params = types.CompletionParams(
        position=types.Position(line=0, character=6),
        text_document=types.TextDocumentIdentifier(uri=str(uri)),
    )
    document = TextDocument(uri=str(uri), source="some yy xx text")
    client_capabilities = types.ClientCapabilities(
        text_document=types.TextDocumentClientCapabilities(
            completion=types.CompletionClientCapabilities(
                context_support=True,
            ),
        ),
    )

    trigger = server.CompletionTrigger(
        patterns=[re.compile("some.*xx"), re.compile("yy.*text")]
    )
    result = trigger(uri, params, document, "rst", client_capabilities)

    assert result is not None
    assert result.match.group(0) == "some yy xx"

    trigger = server.CompletionTrigger(
        patterns=[re.compile("yy.*text"), re.compile("some.*xx")]
    )
    result = trigger(uri, params, document, "rst", client_capabilities)

    assert result is not None
    assert result.match.group(0) == "yy xx text"



from functools import partial

from lsprotocol import types

from esbonio.server import EsbonioLanguageServer
from esbonio.server.features.project_manager import ProjectManager
from esbonio.server.features.sphinx_manager import SphinxClient
from esbonio.server.features.sphinx_manager import SphinxManager


async def refresh_diagnostics(
    server: EsbonioLanguageServer,
    projects: ProjectManager,
    client: SphinxClient,
    *args,
):
    """Refresh sphinx diagnostics."""
    if (project := projects.get_project(client.src_uri)) is None:
        return

    # TODO: Per-client id.
    server.clear_diagnostics("sphinx")

    collection = await project.get_diagnostics()
    for uri, items in collection.items():
        diagnostics = [
            server.converter.structure(item, types.Diagnostic) for item in items
        ]
        server.set_diagnostics("sphinx", uri, diagnostics)

    server.sync_diagnostics()


def esbonio_setup(
    server: EsbonioLanguageServer,
    sphinx_manager: SphinxManager,
    project_manager: ProjectManager,
):
    sphinx_manager.add_listener(
        "app-created", partial(refresh_diagnostics, server, project_manager)
    )
    sphinx_manager.add_listener(
        "build", partial(refresh_diagnostics, server, project_manager)
    )



"""Helper functions for completion support."""

from __future__ import annotations

import re
import typing

from lsprotocol import types

from esbonio import server

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from esbonio.sphinx_agent.types import Role

    RoleRenderer = Callable[
        [server.CompletionContext, Role], types.CompletionItem | None
    ]

    RoleTargetRenderer = Callable[
        [server.CompletionContext, types.CompletionItem], types.CompletionItem | None
    ]


WORD = re.compile("[a-zA-Z]+")
_ROLE_RENDERERS: dict[tuple[str, str], RoleRenderer] = {}
"""CompletionItem rendering functions for roles."""

_ROLE_TARGET_RENDERERS: dict[tuple[str, str], RoleTargetRenderer] = {}
"""CompletionItem rendering functions for role targets."""


def role_renderer(*, language: str, insert_behavior: str):
    """Define a new rendering function for roles."""

    def fn(f: RoleRenderer) -> RoleRenderer:
        _ROLE_RENDERERS[(language, insert_behavior)] = f
        return f

    return fn


def role_target_renderer(*, language: str, insert_behavior: str):
    """Define a new rendering function for role targets."""

    def fn(f: RoleTargetRenderer) -> RoleTargetRenderer:
        _ROLE_TARGET_RENDERERS[(language, insert_behavior)] = f
        return f

    return fn


def get_role_renderer(language: str, insert_behavior: str) -> RoleRenderer | None:
    """Return the role renderer to use.

    Parameters
    ----------
    language
       The source language the completion item will be inserted into

    insert_behavior
       How the completion should behave when inserted.

    Returns
    -------
    Optional[RoleRenderer]
       The rendering function to use that matches the given criteria, if available.
    """
    return _ROLE_RENDERERS.get((language, insert_behavior), None)


def get_role_target_renderer(
    language: str, insert_behavior: str
) -> RoleTargetRenderer | None:
    """Return the role target renderer to use.

    Parameters
    ----------
    language
       The source language the completion item will be inserted into

    insert_behavior
       How the completion should behave when inserted.

    Returns
    -------
    Optional[RoleTargetRenderer]
       The rendering function to use that matches the given criteria, if available.
    """
    return _ROLE_TARGET_RENDERERS.get((language, insert_behavior), None)


@role_renderer(language="rst", insert_behavior="insert")
def render_rst_role_with_insert_text(
    context: server.CompletionContext, role: Role
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``insertText``.

    This implements the ``insert`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    role
       The role.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """

    insert_text = f":{role.name}:"
    user_text = context.match.group(0).strip()

    # Since we can't replace any existing text, it only makes sense
    # to offer completions that align with what the user has already written.
    if not insert_text.startswith(user_text):
        return None

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", " ")):
        start_index = len(user_text)

    # Look for groups of word chars, replace text until the start of the final group
    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_role_common(role)
    item.insert_text = insert_text[start_index:]
    item.insert_text_format = types.InsertTextFormat.PlainText
    return item


@role_target_renderer(language="rst", insert_behavior="replace")
def render_rst_target_with_text_edit(
    context: server.CompletionContext, item: types.CompletionItem
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``insertText``.

    This implements the ``replace`` insert behavior for role targets.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    item
       The ``CompletionItem`` representing the role target.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the item should be skipped
    """
    return item


@role_renderer(language="markdown", insert_behavior="insert")
def render_myst_role_with_insert_text(
    context: server.CompletionContext, role: Role
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``insertText``.

    This implements the ``insert`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    role
       The role.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """

    insert_text = "".join(["{", role.name, "}"])
    user_text = context.match.group(0).strip()

    # Since we can't replace any existing text, it only makes sense
    # to offer completions that align with what the user has already written.
    if not insert_text.startswith(user_text):
        return None

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", "{", "}", "`", " ")):
        start_index = len(user_text)

    # Look for groups of word chars, replace text until the start of the final group
    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_role_common(role)
    item.insert_text = insert_text[start_index:]
    item.insert_text_format = types.InsertTextFormat.PlainText
    return item


@role_renderer(language="rst", insert_behavior="replace")
def render_rst_role_with_text_edit(
    context: server.CompletionContext, role: Role
) -> types.CompletionItem | None:
    """Render a role's ``CompletionItem`` using ``textEdit``.

    This implements the ``replace`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    role
       The role.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """
    match = context.match
    groups = match.groupdict()

    # Insert text starting from the starting ':' character of the role.
    start = match.span()[0] + match.group(0).find(":")
    end = start + len(groups["role"])

    range_ = types.Range(
        start=types.Position(line=context.position.line, character=start),
        end=types.Position(line=context.position.line, character=end),
    )

    insert_text = f":{role.name}:"

    item = _render_role_common(role)
    item.filter_text = insert_text
    item.insert_text_format = types.InsertTextFormat.PlainText
    item.text_edit = types.TextEdit(range=range_, new_text=insert_text)

    return item


@role_renderer(language="markdown", insert_behavior="replace")
def render_myst_role_with_text_edit(
    context: server.CompletionContext, role: Role
) -> types.CompletionItem | None:
    """Render a role's ``CompletionItem`` using ``textEdit``.

    This implements the ``replace`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    role
       The role.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the directive should be skipped
    """
    match = context.match
    groups = match.groupdict()

    # Insert text starting from the starting '{' character of the role.
    start = match.span()[0] + match.group(0).find("{")
    end = start + len(groups["role"])

    range_ = types.Range(
        start=types.Position(line=context.position.line, character=start),
        end=types.Position(line=context.position.line, character=end),
    )

    insert_text = "".join(["{", role.name, "}"])

    item = _render_role_common(role)
    item.filter_text = insert_text
    item.insert_text_format = types.InsertTextFormat.PlainText
    item.text_edit = types.TextEdit(range=range_, new_text=insert_text)

    return item


@role_target_renderer(language="markdown", insert_behavior="replace")
def render_myst_target_with_text_edit(
    context: server.CompletionContext, item: types.CompletionItem
) -> types.CompletionItem | None:
    """Render a ``CompletionItem`` using ``textEdit``.

    This implements the ``replace`` insert behavior for role targets.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    item
       The ``CompletionItem`` representing the role target.

    Returns
    -------
    Optional[types.CompletionItem]
       The rendered completion item, or ``None`` if the item should be skipped
    """
    return item


def _render_role_common(role: Role) -> types.CompletionItem:
    """Render the common fields of a role's completion item."""
    return types.CompletionItem(
        label=role.name,
        detail=role.implementation,
        kind=types.CompletionItemKind.Function,
        data={"completion_type": "role"},
    )



from __future__ import annotations

import typing

from ..app import Database
from ..app import Sphinx
from ..app import logger
from ..types import Uri
from ..util import as_json

if typing.TYPE_CHECKING:
    from sphinx.config import Config


FILES_TABLE = Database.Table(
    "files",
    [
        Database.Column(name="uri", dtype="TEXT"),
        Database.Column(name="docname", dtype="TEXT"),
        Database.Column(name="urlpath", dtype="TEXT"),
    ],
)

CONFIG_TABLE = Database.Table(
    "config",
    [
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="value", dtype="TEXT"),
    ],
)

IGNORED_CONFIG_NAMES = {
    # Declared private by Sphinx in `68af0ac2b`
    "_options",
    # Deprecated/removed in v3.5 and the backwards compatibility code causes issues when
    # dumping the config
    "html_add_permalinks",
}


def init_db(app: Sphinx, config: Config):
    app.esbonio.db.ensure_table(FILES_TABLE)
    app.esbonio.db.ensure_table(CONFIG_TABLE)


def dump_config(app: Sphinx, *args):
    """Dump the user's config into the db so that the parent language server can inspect
    it."""
    app.esbonio.db.clear_table(CONFIG_TABLE)

    values: list[tuple[str, str]] = []
    config = app.config.__getstate__()

    for name, item in config.items():
        if name in IGNORED_CONFIG_NAMES:
            continue

        try:
            values.append(
                (name, as_json(item)),
            )
        except Exception as exc:
            logger.debug(f"Unable to dump config value: {name!r}: {exc}")

    app.esbonio.db.insert_values(CONFIG_TABLE, values)


def build_file_mapping(app: Sphinx, exc: Exception | None):
    """Given a Sphinx application, return a mapping of all known source files to their
    corresponding output files."""

    env = app.env
    builder = app.builder
    files: list[tuple[str, str, str]] = []

    for docname in env.found_docs:
        uri = Uri.for_file(env.doc2path(docname)).resolve()
        build_uri = builder.get_target_uri(docname)

        files.append((str(uri), docname, build_uri))

    # Don't forget any included files.
    # TODO: How best to handle files included in multiple documents?
    for parent_doc, included_docs in env.included.items():
        for docname in included_docs:
            uri = Uri.for_file(env.doc2path(docname)).resolve()
            build_uri = builder.get_target_uri(parent_doc)

            files.append((str(uri), docname, build_uri))

    app.esbonio.db.clear_table(FILES_TABLE)
    app.esbonio.db.insert_values(FILES_TABLE, files)


def setup(app: Sphinx):
    app.connect("config-inited", init_db)
    app.connect("builder-inited", dump_config)
    app.connect("build-finished", build_file_mapping)



from __future__ import annotations

import typing

import pytest
from lsprotocol import types
from pygls.workspace import TextDocument
from pytest_lsp import client_capabilities

from esbonio import server
from esbonio.server.features.directives import completion
from esbonio.server.testing import range_from_str
from esbonio.sphinx_agent.types import MYST_DIRECTIVE
from esbonio.sphinx_agent.types import RST_DIRECTIVE
from esbonio.sphinx_agent.types import Directive

if typing.TYPE_CHECKING:
    from typing import Literal


VSCODE = "visual-studio-code"
NVIM = "neovim@0.10.0"
PATTERNS = {"rst": RST_DIRECTIVE, "markdown": MYST_DIRECTIVE}


@pytest.mark.parametrize(
    "client,language,insert_behavior,directive,text,character,expected",
    [
        (  # Standard rst directive completion with replacement
            VSCODE,
            "rst",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "..",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text=".. image::",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:2"), new_text=".. image::"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # We need to take into account indentation
            VSCODE,
            "rst",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "   ..",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text=".. image::",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:3-0:5"), new_text=".. image::"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # The replacement should remove all exiting text
            VSCODE,
            "rst",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            ".. inc",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text=".. image::",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:6"), new_text=".. image::"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # Unless the directive has an existing argument, then we should not replace
            # it
            VSCODE,
            "rst",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            ".. include:: filename.png",
            3,  # character
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text=".. image::",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:12"), new_text=".. image::"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # When doing 'insert' completions, we should skip directives with
            # incompatible prefixes
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            ".. inc",
            None,
            None,
        ),
        (  # 'insert' completions still do some replacement, but it's up to the client's
            # interpretation of what is considered a 'word'
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            ".. im",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                insert_text="image::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            ".. co",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="code-block::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            ".. code-",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="block::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            ".. code-bl",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="block::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            "..",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text=" c:function::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            ".. c",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="c:function::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            ".. c:",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="function::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            ".. c:fun",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="function::",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (  # If the client supports it, we can use snippets to include the closing fences
            VSCODE,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```""",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:3"), new_text="```{image} $0\n```"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # Otherwise, they should be omitted
            NVIM,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```""",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image}",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:3"), new_text="```{image}"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # Be sure to take into account indentation
            VSCODE,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """   ```""",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                text_edit=types.TextEdit(
                    range=range_from_str("0:3-0:6"), new_text="```{image} $0\n```"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```{inc""",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:7"), new_text="```{image} $0\n```"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```{inc""",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image}",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:7"), new_text="```{image}"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (  # Arguments should not be replaced
            VSCODE,
            "markdown",
            "replace",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```{include} filename.png""",
            7,  # character index
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                filter_text="```{image}",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:12"), new_text="```{image}"
                ),
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "markdown",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            """```{inc""",
            None,
            None,
        ),
        (
            VSCODE,
            "markdown",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "```",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                insert_text="{image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "markdown",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "```{",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                insert_text="image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            VSCODE,
            "markdown",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "```{im",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                insert_text="image} $0\n```",
                insert_text_format=types.InsertTextFormat.Snippet,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="image",
                implementation="docutils.parsers.rst.directives.images.Image",
            ),
            "```{im",
            None,
            types.CompletionItem(
                label="image",
                detail="docutils.parsers.rst.directives.images.Image",
                kind=types.CompletionItemKind.Class,
                insert_text="image}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            "```{co",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="code-block}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            "```{code-",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="block}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="code-block",
                implementation="sphinx.directives.code.CodeBlock",
            ),
            "```{code-bl",
            None,
            types.CompletionItem(
                label="code-block",
                detail="sphinx.directives.code.CodeBlock",
                kind=types.CompletionItemKind.Class,
                insert_text="block}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            "```",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="{c:function}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            "```{c",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="c:function}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            "```{c:",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="function}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
        (
            NVIM,
            "markdown",
            "insert",
            Directive(
                name="c:function",
                implementation="sphinx.domains.c.CFunctionObject",
            ),
            "```{c:fun",
            None,
            types.CompletionItem(
                label="c:function",
                detail="sphinx.domains.c.CFunctionObject",
                kind=types.CompletionItemKind.Class,
                insert_text="function}",
                insert_text_format=types.InsertTextFormat.PlainText,
                data=dict(completion_type="directive"),
            ),
        ),
    ],
)
def test_render_directive_completion(
    client: str,
    language: Literal["rst", "markdown"],
    insert_behavior: Literal["insert", "replace"],
    directive: Directive,
    text: str,
    character: int | None,
    expected: types.CompletionItem | None,
):
    """Ensure that we can render directive completions correctly.

    Parameters
    ----------
    client
       The name of the client to use.

       This will be passed to the ``client_capabilities`` function from ``pytest_lsp``
       and will control what capabilities (e.g. snippet support) will be available.

    language
       The language in which the completion item will be inserted.

    insert_behavior
       How the completion item should behave when inserted.

    directive
       The directive to generate the completion item for.

    text
       The text used to help generate the completion context.

    character
       The character column at which the request is being made.
       If ``None``, it will be assumed that the request is being made at
       the end of ``text``.

    expected
       The expected result.
    """

    match = PATTERNS[language].match(text)
    if not match:
        raise ValueError(f"'{text}' is not valid in this context")

    line = 0
    character = len(text) if character is None else character
    uri = "file:///test.txt"

    context = server.CompletionContext(
        uri=server.Uri.parse(uri),
        doc=TextDocument(uri=uri),
        match=match,
        position=types.Position(line=line, character=character),
        language=language,
        capabilities=client_capabilities(client),
    )

    render_func = completion.get_directive_renderer(language, insert_behavior)
    assert render_func is not None

    item = render_func(context, directive)
    if expected is None:
        assert item is None
    else:
        assert item == expected



from __future__ import annotations

import pathlib
from typing import Any

import pytest
from pygls.protocol import default_converter

from esbonio.server import Uri
from esbonio.server.features.project_manager import Project
from esbonio.server.features.sphinx_manager.client_subprocess import (
    SubprocessSphinxClient,
)
from esbonio.sphinx_agent import types


def check_diagnostics(
    expected: dict[Uri, list[types.Diagnostic]],
    actual: dict[Uri, list[dict[str, Any]]],
):
    """Ensure that two sets of diagnostics are equal."""
    converter = default_converter()
    assert set(actual.keys()) == set(expected.keys())

    for k, ex_diags in expected.items():
        actual_diags = [converter.structure(d, types.Diagnostic) for d in actual[k]]

        assert len(ex_diags) == len(actual_diags)

        for actual_diagnostic in actual_diags:
            # Assumes ranges are unique
            matches = [e for e in ex_diags if e.range == actual_diagnostic.range]
            assert len(matches) == 1

            expected_diagnostic = matches[0]
            assert actual_diagnostic.range == expected_diagnostic.range
            assert actual_diagnostic.severity == expected_diagnostic.severity
            assert actual_diagnostic.message.startswith(expected_diagnostic.message)


@pytest.mark.asyncio
async def test_diagnostics(client: SubprocessSphinxClient, project: Project, uri_for):
    """Ensure that the sphinx agent reports diagnostics collected during the build, and
    that they are correctly reset when fixed."""
    rst_diagnostics_uri = uri_for("workspaces/demo/rst/diagnostics.rst")
    myst_diagnostics_uri = uri_for("workspaces/demo/myst/diagnostics.md")
    index_uri = uri_for("workspaces/demo/index.rst")
    conf_uri = uri_for("workspaces/demo/conf.py")

    message = "undefined label: 'not-a-real-reference'"

    expected = {
        index_uri: [
            types.Diagnostic(
                message='Unknown directive type "grid"',
                severity=types.DiagnosticSeverity.Error,
                range=types.Range(
                    start=types.Position(line=13, character=0),
                    end=types.Position(line=14, character=0),
                ),
            )
        ],
        conf_uri: [
            types.Diagnostic(
                message="Could not import extension sphinx_design",
                severity=types.DiagnosticSeverity.Error,
                range=types.Range(
                    start=types.Position(line=20, character=4),
                    end=types.Position(line=20, character=19),
                ),
            ),
            types.Diagnostic(
                message="no theme named 'furo' found",
                severity=types.DiagnosticSeverity.Error,
                range=types.Range(
                    start=types.Position(line=42, character=0),
                    end=types.Position(line=42, character=19),
                ),
            ),
        ],
        rst_diagnostics_uri: [
            types.Diagnostic(
                message=message,
                severity=types.DiagnosticSeverity.Warning,
                range=types.Range(
                    start=types.Position(line=5, character=0),
                    end=types.Position(line=6, character=0),
                ),
            ),
        ],
        myst_diagnostics_uri: [
            types.Diagnostic(
                message=message,
                severity=types.DiagnosticSeverity.Warning,
                range=types.Range(
                    start=types.Position(line=4, character=0),
                    end=types.Position(line=5, character=0),
                ),
            ),
        ],
    }

    actual = await project.get_diagnostics()
    check_diagnostics(expected, actual)

    await client.build(
        content_overrides={
            str(
                rst_diagnostics_uri
            ): "My Custom Title\n===============\n\nThere are no images here"
        }
    )

    actual = await project.get_diagnostics()

    fixed_expected = expected.copy()
    del fixed_expected[rst_diagnostics_uri]
    check_diagnostics(fixed_expected, actual)

    # The original diagnostics should be reported when the issues are re-introduced.
    #
    # Note: We have to "override" the contents of the file with the original text to
    #       trick Sphinx into re-building the file.
    await client.build(
        content_overrides={
            str(rst_diagnostics_uri): pathlib.Path(rst_diagnostics_uri).read_text()
        }
    )
    actual = await project.get_diagnostics()
    check_diagnostics(expected, actual)
